@@ -7,14 +7,14 @@ import numpy as np
 from .tensors import volume, create_ndarray_f32, create_ndarray_int64, get_attribute_data, is_valid_ndarray
 from .utils import NODEPROFILER_REGISTRY, tuple2str
 
-ADD_MACS = 0.5
+ADD_MACS = 1
 EXP_MACS = 16
 LOG_MACS = 16
 SQRT_MACS = 4
 POW_MACS = 32
 MUL_MACS = 1
 DIV_MACS = 2
-CMP_MACS = 0.2
+CMP_MACS = 1
 SIN_MACS = 14
 COS_MACS = 14
 
@@ -1029,14 +1029,16 @@ class Constant(NodeBase):
 class Unsqueeze(NodeBase):
     def __init__(self, nodeproto):
         super().__init__(nodeproto)
-        self.axes = [0]
-        for att in nodeproto.attribute:
-            if att.name == 'axes':
-                self.axes = get_attribute_data(att)
+        self.axes = None
+        auto_add_attributes(nodeproto.attribute, ['axes'], self)
 
     def infer_shape(self, intensors: [numpy.ndarray]):
         outtensor = intensors[0]
-        for axis in self.axes:
+        if self.axes is None:
+            axes = intensors[1]
+        else:
+            axes = self.axes
+        for axis in axes:
             outtensor = numpy.expand_dims(outtensor, axis=axis)
         return [outtensor]
 
@@ -1199,7 +1201,11 @@ class Split(NodeBase):
         split = []
         end = 0
         if self.split is None:
-            self.split = [intensors[0].shape[self.axis] // 2]
+            if len(intensors) == 2:
+                self.split = intensors[1]
+            else:
+                self.split = [intensors[0].shape[self.axis] // 2]
+
         self.axis = axes_neg2pos(len(intensors[0].shape), [self.axis])[0]
         for v in self.split:
             if end + v >= intensors[0].shape[self.axis]:
@@ -1487,6 +1493,150 @@ class LeakyRelu(PWNBase):
         self.op_mac = MUL_MACS + CMP_MACS
 
 
+@NODEPROFILER_REGISTRY.register()
+class Einsum(NodeBase):
+    def __init__(self, node_proto):
+        super().__init__(node_proto)
+        auto_add_attributes(node_proto.attribute, ['equation'], self)
+        strs = self.equation.split(b',')
+        self.ashape = strs[0].replace(b' ', b'')
+        strs = strs[1].split(b'->')
+        self.bshape = strs[0].replace(b' ', b'')
+        self.cshape = strs[1].replace(b' ', b'')
+
+    def infer_shape(self, intensors: [numpy.ndarray]):
+        shape = []
+        map = {}
+        for i, v in enumerate(intensors[0].shape):
+            map[self.ashape[i]] = v
+        for i, v in enumerate(intensors[1].shape):
+            map[self.bshape[i]] = v
+        for k in self.cshape:
+            shape.append(map[k])
+        outtensor = create_ndarray_f32(shape)
+        return [outtensor]
+
+    def profile(self, intensors: [numpy.ndarray], outtensors: [numpy.ndarray]):
+        macs = 1
+        params = 0
+        map = {}
+        for i, v in enumerate(intensors[0].shape):
+            map[self.ashape[i]] = v
+        for i, v in enumerate(intensors[1].shape):
+            map[self.bshape[i]] = v
+        for key in map.keys():
+            macs *= map[key]
+        return macs, params
+
+
+@NODEPROFILER_REGISTRY.register()
+class DequantizeLinear(PWNBase):
+    def __init__(self, node_proto):
+        super().__init__(node_proto)
+        self.op_mac = MUL_MACS
+
+
+@NODEPROFILER_REGISTRY.register()
+class QuantizeLinear(PWNBase):
+    def __init__(self, node_proto):
+        super().__init__(node_proto)
+        self.op_mac = MUL_MACS
+
+
+@NODEPROFILER_REGISTRY.register()
+class MatMulInteger(Gemm):
+    pass
+
+
+@NODEPROFILER_REGISTRY.register()
+class QLinearMatMul(Gemm):
+    def __init__(self, nodeproto):
+        super().__init__(nodeproto)
+        attnames = ['transA', 'transB']
+        self.transA = None
+        self.transB = None
+        auto_add_attributes(nodeproto.attribute, attnames, self)
+
+    def infer_shape(self, intensors: [numpy.ndarray]):
+        x = intensors[0]
+        w = intensors[3]
+
+        if self.__class__ == Gemm:
+            if self.transA is not None and self.transA > 0:
+                xshape = x.shape[::-1]
+            else:
+                xshape = x.shape
+            if self.transB is not None and self.transB > 0:
+                yshape = xshape[:-1] + (w.shape[-2],)
+            else:
+                yshape = xshape[:-1] + (w.shape[-1],)
+        else:
+            yshape = x.shape[:-1] + (w.shape[-1],)
+
+        return [create_ndarray_f32(yshape)]
+
+    def profile(self, intensors: [numpy.ndarray], outtensors: [numpy.ndarray]):
+        params = 0
+        xshape = intensors[0].shape
+        weight_shape = intensors[3].shape
+        macs = volume(xshape)
+        if self.__class__ == Gemm:
+            macs *= weight_shape[0]
+        else:
+            macs *= weight_shape[-1]
+        return macs, params
+
+
+@NODEPROFILER_REGISTRY.register()
+class QLinearConv(Conv):
+    def infer_shape(self, intensors: [numpy.ndarray]):
+        outtensors = []
+        xtensor = intensors[0]
+        wtensor = intensors[3]
+        xshape = xtensor.shape
+        wshape = wtensor.shape
+        shape = []
+        if self.auto_pad is not None and self.auto_pad != b'NOTSET':
+            if self.auto_pad in [b'SAME_LOWER', b'SAME_UPPER']:
+                shape = (xshape[0], wshape[0], math.ceil(xshape[2] / self.strides[0]))
+                if len(xshape) == 4:
+                    shape += (math.ceil(xshape[3] / self.strides[1]),)
+        else:
+            if len(xshape) == 4:
+                oh = conv_output_shape(xshape[2], self.pads[0] + self.pads[2], wshape[2], self.strides[0],
+                                       self.dilations[0])
+                ow = conv_output_shape(xshape[3], self.pads[1] + self.pads[3], wshape[3], self.strides[1],
+                                       self.dilations[1])
+                shape = (xshape[0], wshape[0], oh, ow)
+            elif len(xshape) == 3:
+                oh = conv_output_shape(xshape[2], self.pads[0] + self.pads[1], wshape[2], self.strides[0],
+                                       self.dilations[0])
+                shape = (xshape[0], wshape[0], oh)
+        outtensors.append(create_ndarray_f32(shape))
+        return outtensors
+
+    def profile(self, intensors: [numpy.ndarray], outtensors: [numpy.ndarray]):
+        macs = 0
+        params = 0
+        if self.nboutput == 1:
+            kernel_shape = intensors[3].shape
+            params += volume(kernel_shape)
+            if self.nbinput == 3:
+                params += kernel_shape[0]
+
+            if len(kernel_shape) > 3:
+                outvol = volume(outtensors[0].shape)
+                macs += outvol * kernel_shape[1] * kernel_shape[2] * kernel_shape[3]
+            elif len(kernel_shape) == 3:
+                outvol = volume(outtensors[0].shape)
+                macs += outvol * kernel_shape[1] * kernel_shape[2]
+
+            if self.nbinput == 9:
+                macs += (outvol * ADD_MACS)
+
+        return macs, params
+
+
 def node_profile(node_proto: str, ins: [], outs: []):
     node_class = NODEPROFILER_REGISTRY.get(node_proto.op_type)
     if node_class != None:
